from __future__ import annotations

from typing import Optional, cast

from tomlkit.items import Array, Table
from tomlkit.toml_document import TOMLDocument

from .config import Config
from .pep508 import normalize_pep508_array
from .util import ensure_newline_at_end, order_keys, sorted_array


def fmt_build_system(parsed: TOMLDocument, conf: Config) -> None:
    system = cast(Optional[Table], parsed.get("build-system"))
    if system is not None:
        normalize_pep508_array(cast(Optional[Array], system.get("requires")), conf.indent)
        sorted_array(cast(Optional[Array], system.get("backend-path")), indent=conf.indent)
        order_keys(system, ("build-backend", "requires", "backend-path"))
        ensure_newline_at_end(system)


__all__ = [
    "fmt_build_system",
]



from __future__ import annotations

from typing import cast

from tomlkit import TOMLDocument
from tomlkit.items import Table

from .config import Config
from .util import ensure_newline_at_end, order_keys


def fmt_tools(parsed: TOMLDocument, conf: Config) -> None:  # noqa: U100
    tools: Table | None = parsed.get("tool")
    if tools is None:
        return None
    for tool in tools:
        table = tools[tool]
        ensure_newline_at_end(cast(Table, table))
    order = [
        "setuptools",
        "setuptools_scm",
        "hatch",
        "black",
        "isort",
        "flake8",
        "pytest",
        "coverage",
        "mypy",
    ]
    order_keys(tools, to_pin=order)


__all__ = [
    "fmt_tools",
]



from __future__ import annotations

import sys
from collections import defaultdict
from dataclasses import dataclass, field
from typing import Any, Callable, Sequence

from tomlkit.container import OutOfOrderTableProxy
from tomlkit.items import (
    AbstractTable,
    Array,
    Comment,
    Item,
    Key,
    String,
    Table,
    Trivia,
    Whitespace,
    _ArrayItemGroup,
)

if sys.version_info >= (3, 8):  # pragma: no cover (py38+)
    from typing import Protocol
else:  # pragma: no cover (<py38)
    from typing_extensions import Protocol


class SupportsDunderLT(Protocol):
    def __lt__(self, __other: Any) -> bool:  # noqa: U101
        ...


class SupportsDunderGT(Protocol):
    def __gt__(self, __other: Any) -> bool:  # noqa: U101
        ...


def sort_inline_table(item: tuple[str, Any | Table]) -> str:
    key, value = item
    return f"{key}{'-'.join(value) if isinstance(value, Table) else ''}"


def order_keys(
    table: AbstractTable | OutOfOrderTableProxy,
    to_pin: Sequence[str] | None = None,
    sort_key: None | Callable[[tuple[str, tuple[Key, Item]]], SupportsDunderLT | SupportsDunderGT] = None,
) -> None:
    if isinstance(table, OutOfOrderTableProxy):
        return  # pragma: no cover
    body = table.value.body
    entries: dict[str, list[Any]] = defaultdict(list)
    for key, value in body:
        if isinstance(key, Key):
            entries[key.key].append((key, value))
    body.clear()

    for pin in to_pin or []:  # push pinned to start
        if pin in entries:
            body.extend(sorted(entries[pin], key=sort_inline_table))  # type: ignore
            del entries[pin]
    # append the rest
    if sort_key is None:
        for elements in entries.values():
            body.extend(elements)  # pragma: no cover
    else:
        for _, elements in sorted(entries.items(), key=sort_inline_table):
            body.extend(elements)


@dataclass
class ArrayEntries:
    text: String
    comments: list[Comment] = field(default_factory=list)


def sorted_array(
    array: Array | None, indent: int, key: Callable[[ArrayEntries], str] = lambda e: str(e.text).lower()
) -> None:
    if array is None:
        return
    body = array._value

    entries: list[ArrayEntries] = []
    start: list[Comment] = []
    for group in body:
        for entry in group:
            if isinstance(entry, String):
                entries.append(ArrayEntries(entry))
            elif isinstance(entry, Comment):
                (entries[-1].comments if len(entries) else start).append(entry)

    body.clear()
    indent_text = " " * indent
    for start_entry in start:
        body.append(_ArrayItemGroup(indent=Whitespace(f"\n{indent_text}"), comment=start_entry))
    for element in sorted(entries, key=key):
        if element.comments:
            com = " ".join(i.trivia.comment[1:].strip() for i in element.comments)
            comment = Comment(Trivia(comment=f" # {com}", trail=""))
        else:
            comment = None
        group = _ArrayItemGroup(
            indent=Whitespace(f"\n{indent_text}"), value=element.text, comma=Whitespace(","), comment=comment
        )
        body.append(group)
    body.append(_ArrayItemGroup(indent=Whitespace("\n")))


def ensure_newline_at_end(body: Table) -> None:
    content = body
    while content.value.body and isinstance(content.value.body[-1][1], Table):
        content = content.value.body[-1][1]
    whitespace = Whitespace("\n")
    insert_body = content.value.body
    if insert_body and isinstance(insert_body[-1][1], Whitespace):
        insert_body[-1] = insert_body[-1][0], whitespace
    else:
        insert_body.append((None, whitespace))


__all__ = [
    "ArrayEntries",
    "sorted_array",
    "order_keys",
    "ensure_newline_at_end",
]



from __future__ import annotations

from pathlib import Path
from textwrap import dedent
from typing import Callable

import pytest
from pytest_mock import MockerFixture
from tomlkit.toml_document import TOMLDocument

from pyproject_fmt.formatter import format_pyproject
from pyproject_fmt.formatter.config import Config
from tests import Fmt


@pytest.fixture()
def fmt(mocker: MockerFixture) -> Fmt:
    def _func(formatter: Callable[[TOMLDocument, Config], None], start: str, expected: str) -> None:
        mocker.patch("pyproject_fmt.formatter._perform", formatter)
        opts = Config(pyproject_toml=Path(), toml=dedent(start))
        result = format_pyproject(opts)

        expected = dedent(expected)
        assert result == expected

    return _func



from __future__ import annotations

import pytest

from pyproject_fmt.formatter.project import fmt_project
from tests import Fmt


@pytest.mark.parametrize(
    "value",
    [
        "[project]\nname='a-b'",
        "[project]\nname='A_B'",
        "[project]\nname='a.-..-__B'",
    ],
)
def test_project_name(fmt: Fmt, value: str) -> None:
    fmt(fmt_project, value, '[project]\nname="a-b"\n')


def test_project_dependencies(fmt: Fmt) -> None:
    start = '[project]\ndependencies=["pytest","pytest-cov",]'
    expected = '[project]\ndependencies=[\n  "pytest",\n  "pytest-cov",\n]\n'
    fmt(fmt_project, start, expected)


def test_project_dependencies_with_double_quotes(fmt: Fmt) -> None:
    start = """
    [project]
    dependencies = [
        'packaging>=20.0;python_version>"3.4"',
        "appdirs"
    ]
    """
    expected = """
    [project]
    dependencies = [
      "appdirs",
      'packaging>=20; python_version > "3.4"',
    ]
    """
    fmt(fmt_project, start, expected)


def test_project_dependencies_with_mixed_quotes(fmt: Fmt) -> None:
    start = """
    [project]
    dependencies = [
        "packaging>=20.0;python_version>\\"3.4\\" and python_version != '3.5'",
        "foobar@ git+https://weird-vcs/w/index.php?param=org'repo ; python_version == '2.7'",
        "appdirs"
    ]
    """
    expected = """
    [project]
    dependencies = [
      "appdirs",
      "foobar@ git+https://weird-vcs/w/index.php?param=org'repo ; python_version == \\"2.7\\"",
      'packaging>=20; python_version > "3.4" and python_version != "3.5"',
    ]
    """
    fmt(fmt_project, start, expected)


def test_project_description(fmt: Fmt) -> None:
    start = '[project]\ndescription=" Magical stuff\t"'
    expected = '[project]\ndescription="Magical stuff"\n'
    fmt(fmt_project, start, expected)


def test_project_scripts(fmt: Fmt) -> None:
    start = """
    [project.scripts]
    c = "d"
    a = "b"
    """
    expected = """
    [project.scripts]
    a = "b"
    c = "d"
    """
    fmt(fmt_project, start, expected)


def test_project_optional_dependencies(fmt: Fmt) -> None:
    start = """
    [project.optional-dependencies]
    test = ["B", "A"]
    docs = [ "C",
    "D"]
    """
    expected = """
    [project.optional-dependencies]
    docs = [
      "C",
      "D",
    ]
    test = [
      "A",
      "B",
    ]
    """
    fmt(fmt_project, start, expected)


def test_entry_points(fmt: Fmt) -> None:
    start = """
    [project.entry-points]
    beta = {C = "c", D = "d"}
    alpha = {B = "b", "A.A" = "a"}
    """
    expected = """
    [project.entry-points]
    alpha = {"A.A" = "a",B = "b"}
    beta = {C = "c",D = "d"}
    """
    fmt(fmt_project, start, expected)



from __future__ import annotations

import os
import sys
from pathlib import Path
from stat import S_IREAD, S_IWRITE

import pytest

from pyproject_fmt.cli import cli_args


def test_cli_pyproject_toml_ok(tmp_path: Path) -> None:
    path = tmp_path / "tox.ini"
    path.write_text("")
    result = cli_args([str(path)])
    assert result.inputs == [path]


def test_cli_inputs_ok(tmp_path: Path) -> None:
    paths = []
    for filename in ("tox.ini", "tox2.ini", "tox3.ini"):
        path = tmp_path / filename
        path.write_text("")
        paths.append(path)
    result = cli_args([*map(str, paths)])
    assert result.inputs == paths


def test_cli_pyproject_toml_not_exists(tmp_path: Path, capsys: pytest.CaptureFixture[str]) -> None:
    with pytest.raises(SystemExit) as context:
        cli_args([str(tmp_path / "tox.ini")])
    assert context.value.code != 0
    out, err = capsys.readouterr()
    assert not out
    assert "argument inputs: path does not exist" in err


def test_cli_pyproject_toml_not_file(tmp_path: Path, capsys: pytest.CaptureFixture[str]) -> None:
    path = tmp_path / "temp"
    os.mkfifo(path)
    with pytest.raises(SystemExit) as context:
        cli_args([str(path)])
    assert context.value.code != 0
    out, err = capsys.readouterr()
    assert not out
    assert "argument inputs: path is not a file" in err


@pytest.mark.parametrize(("flag", "error"), [(S_IREAD, "write"), (S_IWRITE, "read")])
@pytest.mark.skipif(sys.platform == "win32", reason="On Windows files cannot be read only, only folders")
def test_cli_pyproject_toml_permission_fail(
    tmp_path: Path, capsys: pytest.CaptureFixture[str], flag: int, error: str
) -> None:
    path = tmp_path / "tox.ini"
    path.write_text("")
    path.chmod(flag)
    try:
        with pytest.raises(SystemExit) as context:
            cli_args([str(path)])
    finally:
        path.chmod(S_IWRITE | S_IREAD)
    assert context.value.code != 0
    out, err = capsys.readouterr()
    assert not out
    assert f"argument inputs: cannot {error} path" in err


def test_pyproject_toml_resolved(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.chdir(tmp_path)
    path = tmp_path / "tox.ini"
    path.write_text("")
    result = cli_args(["tox.ini"])
    assert result.inputs == [path]


def test_pyproject_toml_dir(tmp_path: Path) -> None:
    (tmp_path / "pyproject.toml").write_text("")
    cli_args([str(tmp_path)])



from __future__ import annotations

from typing import Optional, cast

from packaging.utils import canonicalize_name
from tomlkit.items import Array, String, Table
from tomlkit.toml_document import TOMLDocument

from .config import Config
from .pep508 import normalize_pep508_array
from .util import ensure_newline_at_end, order_keys, sorted_array


def fmt_project(parsed: TOMLDocument, conf: Config) -> None:
    project = cast(Optional[Table], parsed.get("project"))
    if project is None:
        return

    if "name" in project:  # normalize names to hyphen so sdist / wheel have the same prefix
        name = project["name"]
        assert isinstance(name, str)
        project["name"] = canonicalize_name(name)
    if "description" in project:
        project["description"] = String.from_raw(str(project["description"]).strip())

    sorted_array(cast(Optional[Array], project.get("keywords")), indent=conf.indent)
    sorted_array(cast(Optional[Array], project.get("dynamic")), indent=conf.indent)

    normalize_pep508_array(cast(Optional[Array], project.get("dependencies")), conf.indent)
    if "optional-dependencies" in project:
        opt_deps = cast(Table, project["optional-dependencies"])
        for value in opt_deps.values():
            normalize_pep508_array(cast(Array, value), conf.indent)
        order_keys(opt_deps, (), sort_key=lambda k: k[0])  # pragma: no branch

    for of_type in ("scripts", "gui-scripts", "entry-points", "urls"):
        if of_type in project:
            table = cast(Table, project[of_type])
            order_keys(table, (), sort_key=lambda k: k[0])  # pragma: no branch

    if "entry-points" in project:  # order entry points sub-table
        entry_points = cast(Table, project["entry-points"])
        order_keys(entry_points, (), sort_key=lambda k: k[0])  # pragma: no branch
        for entry_point in entry_points.values():
            order_keys(entry_point, (), sort_key=lambda k: k[0])  # pragma: no branch

    # license: Optional[Union[str, LicenseTableLegacy]]
    # license_files: Optional[LicenseFilesTable] = Field(alias="license-files")
    # readme: Optional[Union[str, ReadmeTable]]
    # order maintainers and authors table
    # update classifiers depending on requires
    # handle readme table

    key_order = ["name", "version", "description", "readme", "keywords", "license", "license-files"]
    key_order.extend(["maintainers", "authors", "requires-python", "classifiers", "dynamic", "dependencies"])
    # these go at the end as they may be inline or exploded
    key_order.extend(["optional-dependencies", "urls", "scripts", "gui-scripts", "entry-points"])
    order_keys(project, key_order)
    ensure_newline_at_end(project)


__all__ = [
    "fmt_project",
]



from __future__ import annotations

import pytest

from pyproject_fmt.formatter.pep508 import normalize_req, normalize_requires


@pytest.mark.parametrize(
    ("value", "result"),
    [
        ("", []),
        ("\t", []),
        ("\n\t\n", []),
        ("b\na\n", ["a", "b"]),
        ("a\nb\n", ["a", "b"]),
        ("A\na\n", ["A", "a"]),
        ("\nA\n\n\nb\n\n", ["A", "b"]),
        (
            'packaging>=20.0;python_version>"3.4"\n'
            "xonsh>=0.9.16;python_version > '3.4' and python_version != '3.9'\n"
            "pytest-xdist>=1.31.0\n"
            "foo@http://foo.com\n"
            "bar [fred,al] @ http://bat.com ;python_version=='2.7'\n"
            "baz [quux, strange];python_version<\"2.7\" and platform_version=='2'\n",
            [
                "foo@ http://foo.com",
                "pytest-xdist>=1.31",
                'bar[al,fred]@ http://bat.com ; python_version == "2.7"',
                'baz[quux,strange]; python_version < "2.7" and platform_version == "2"',
                'packaging>=20; python_version > "3.4"',
                'xonsh>=0.9.16; python_version > "3.4" and python_version != "3.9"',
            ],
        ),
        ("pytest>=6.0.0", ["pytest>=6"]),
        ("pytest==6.0.0", ["pytest==6"]),
        ("pytest~=6.0.0", ["pytest~=6.0.0"]),
    ],
)
def test_requires_fmt(value: str, result: list[str]) -> None:
    outcome = normalize_requires([i.strip() for i in value.splitlines() if i.strip()])
    assert outcome == result


@pytest.mark.parametrize("char", ["!", "=", ">", "<", " ", "\t", "@"])
def test_bad_syntax_requires(char: str) -> None:
    with pytest.raises(ValueError, match=f"[{char}]" if char.strip() else None):
        normalize_req(f"{char};")



from __future__ import annotations

from pyproject_fmt.formatter.tools import fmt_tools
from tests import Fmt


def test_tools_ordering(fmt: Fmt) -> None:
    content = """
    [tool.coverage]
    a = 0
    [tool.pytest]
    a = 0
    [tool.black]
    a = 0
    [tool.isort]
    a = 0
    [tool.flake8]
    [tool.hatch]
    a = 0
    [tool.setuptools_scm]
    a = 0
    [tool.setuptools]
    a.b = 0

    """
    expected = """
    [tool.setuptools]
    a.b = 0

    [tool.setuptools_scm]
    a = 0

    [tool.hatch]
    a = 0

    [tool.black]
    a = 0

    [tool.isort]
    a = 0

    [tool.flake8]

    [tool.pytest]
    a = 0

    [tool.coverage]
    a = 0
        """
    fmt(fmt_tools, content, expected)



from __future__ import annotations

import os
from argparse import (
    ArgumentDefaultsHelpFormatter,
    ArgumentParser,
    ArgumentTypeError,
    Namespace,
)
from pathlib import Path
from typing import Sequence

from pyproject_fmt.formatter.config import DEFAULT_INDENT, Config


class PyProjectFmtNamespace(Namespace):
    """Options for pyproject-fmt tool"""

    inputs: list[Path]
    stdout: bool
    indent: int
    check: bool

    @property
    def configs(self) -> list[Config]:
        return [Config(toml, toml.read_text(encoding="utf-8"), self.indent) for toml in self.inputs]


def pyproject_toml_path_creator(argument: str) -> Path:
    """Validate that pyproject.toml can be formatted.

    :param argument: the string argument passed in
    :return: the pyproject.toml path
    """
    path = Path(argument).absolute()
    if path.is_dir():
        path = path / "pyproject.toml"
    if not path.exists():
        raise ArgumentTypeError("path does not exist")
    if not path.is_file():
        raise ArgumentTypeError("path is not a file")
    if not os.access(path, os.R_OK):
        raise ArgumentTypeError("cannot read path")  # pragma: no cover
    if not os.access(path, os.W_OK):
        raise ArgumentTypeError("cannot write path")  # pragma: no cover
    return path


def _build_cli() -> ArgumentParser:
    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter)
    group = parser.add_mutually_exclusive_group()
    msg = "print the formatted text to the stdout (instead of update in-place)"
    group.add_argument("-s", "--stdout", action="store_true", help=msg)
    msg = "check and fail if any input would be formatted, printing any diffs"
    group.add_argument("--check", action="store_true", help=msg)
    parser.add_argument("--indent", type=int, default=DEFAULT_INDENT, help="number of spaces to indent")
    msg = "pyproject.toml file(s) to format"
    parser.add_argument("inputs", nargs="+", type=pyproject_toml_path_creator, help=msg)
    return parser


def cli_args(args: Sequence[str]) -> PyProjectFmtNamespace:
    """
    Load the tools options.

    :param args: CLI arguments
    :return: the parsed options
    """
    parser = _build_cli()
    return parser.parse_args(namespace=PyProjectFmtNamespace(), args=args)


__all__ = [
    "cli_args",
    "PyProjectFmtNamespace",
]



from __future__ import annotations

import sys
from dataclasses import dataclass
from pathlib import Path

if sys.version_info >= (3, 8):  # pragma: no cover (py38+)
    from typing import Final
else:  # pragma: no cover (<py38)
    from typing_extensions import Final


DEFAULT_INDENT: Final[int] = 2  #: default indentation level


@dataclass(frozen=True)
class Config:
    """Configuration flags for the formatting."""

    pyproject_toml: Path
    toml: str  #: the text to format
    indent: int = DEFAULT_INDENT  #: indentation to apply


__all__ = [
    "Config",
    "DEFAULT_INDENT",
]
